@@ -1,6 +1,6 @@
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2015 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2016 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -61,13 +61,13 @@ dhcp_get_hostname(char *buf, size_t buf_len, const struct if_options *ifo)
 			return NULL;
 		buf[buf_len - 1] = '\0';
 	} else
-		strlcpy(buf, ifo->hostname, sizeof(buf));
+		strlcpy(buf, ifo->hostname, buf_len);
 
 	/* Deny sending of these local hostnames */
-	if (strcmp(buf, "(none)") == 0 ||
+	if (buf[0] == '\0' || buf[0] == '.' ||
+	    strcmp(buf, "(none)") == 0 ||
 	    strcmp(buf, "localhost") == 0 ||
-	    strncmp(buf, "localhost.", strlen("localhost.")) == 0 ||
-	    buf[0] == '.')
+	    strncmp(buf, "localhost.", strlen("localhost.")) == 0)
 		return NULL;
 
 	/* Shorten the hostname if required */
@@ -579,48 +579,76 @@ print_string(char *dst, size_t len, int type, const uint8_t *data, size_t dl)
 	return (ssize_t)bytes;
 }
 
-#define ADDRSZ		4
 #define ADDR6SZ		16
 static ssize_t
 dhcp_optlen(const struct dhcp_opt *opt, size_t dl)
 {
 	size_t sz;
 
-	if (opt->type == 0 ||
-	    opt->type & (STRING | BINHEX | RFC3442))
-	{
-		if (opt->len) {
-			if ((size_t)opt->len > dl)
-				return -1;
-			return (ssize_t)opt->len;
-		}
-		return (ssize_t)dl;
-	}
-
-	if ((opt->type & (ADDRIPV4 | ARRAY)) == (ADDRIPV4 | ARRAY)) {
-		if (dl < ADDRSZ)
-			return -1;
-		return (ssize_t)(dl - (dl % ADDRSZ));
-	}
-
-	if ((opt->type & (ADDRIPV6 | ARRAY)) == (ADDRIPV6 | ARRAY)) {
-		if (dl < ADDR6SZ)
-			return -1;
-		return (ssize_t)(dl - (dl % ADDR6SZ));
-	}
-
-	if (opt->type & (UINT32 | ADDRIPV4))
+	if (opt->type & ADDRIPV6)
+		sz = ADDR6SZ;
+	else if (opt->type & (UINT32 | ADDRIPV4))
 		sz = sizeof(uint32_t);
 	else if (opt->type & UINT16)
 		sz = sizeof(uint16_t);
 	else if (opt->type & (UINT8 | BITFLAG))
 		sz = sizeof(uint8_t);
-	else if (opt->type & ADDRIPV6)
-		sz = ADDR6SZ;
-	else
-		/* If we don't know the size, assume it's valid */
+	else if (opt->type & FLAG)
+		return 0;
+	else {
+		/* All other types are variable length */
+		if (opt->len) {
+			if ((size_t)opt->len > dl) {
+				errno = EOVERFLOW;
+				return -1;
+			}
+			return (ssize_t)opt->len;
+		}
 		return (ssize_t)dl;
-	return dl < sz ? -1 : (ssize_t)sz;
+	}
+	if (dl < sz) {
+		errno = EOVERFLOW;
+		return -1;
+	}
+
+	/* Trim any extra data.
+	 * Maybe we need a settng to reject DHCP options with extra data? */
+	if (opt->type & ARRAY)
+		return (ssize_t)(dl - (dl % sz));
+	return (ssize_t)sz;
+}
+
+/* It's possible for DHCPv4 to contain an IPv6 address */
+static ssize_t
+ipv6_printaddr(char *s, size_t sl, const uint8_t *d, const char *ifname)
+{
+	char buf[INET6_ADDRSTRLEN];
+	const char *p;
+	size_t l;
+
+	p = inet_ntop(AF_INET6, d, buf, sizeof(buf));
+	if (p == NULL)
+		return -1;
+
+	l = strlen(p);
+	if (d[0] == 0xfe && (d[1] & 0xc0) == 0x80)
+		l += 1 + strlen(ifname);
+
+	if (s == NULL)
+		return (ssize_t)l;
+
+	if (sl < l) {
+		errno = ENOMEM;
+		return -1;
+	}
+
+	s += strlcpy(s, p, sl);
+	if (d[0] == 0xfe && (d[1] & 0xc0) == 0x80) {
+		*s++ = '%';
+		s += strlcpy(s, ifname, sl);
+	}
+	*s = '\0';
+	return (ssize_t)l;
 }
 
 static ssize_t
@@ -637,10 +665,6 @@ print_option(char *s, size_t len, const struct dhcp_opt *opt,
 	size_t l;
 	char *tmp;
 
-#ifndef INET6
-	UNUSED(ifname);
-#endif
-
 	if (opt->type & RFC1035) {
 		sl = decode_rfc1035(NULL, 0, data, dl);
 		if (sl == 0 || sl == -1)
@@ -721,23 +745,20 @@ print_option(char *s, size_t len, const struct dhcp_opt *opt,
 		} else if (opt->type & ADDRIPV4) {
 			l = 16;
 			dl /= 4;
-		}
-#ifdef INET6
-		else if (opt->type & ADDRIPV6) {
+		} else if (opt->type & ADDRIPV6) {
 			e = data + dl;
 			l = 0;
 			while (data < e) {
 				if (l)
 					l++; /* space */
 				sl = ipv6_printaddr(NULL, 0, data, ifname);
-				if (sl != -1)
-					l += (size_t)sl;
+				if (sl == -1)
+					return l == 0 ? -1 : (ssize_t)l;
+				l += (size_t)sl;
 				data += 16;
 			}
 			return (ssize_t)l;
-		}
-#endif
-		else {
+		} else {
 			errno = EINVAL;
 			return -1;
 		}
@@ -779,21 +800,15 @@ print_option(char *s, size_t len, const struct dhcp_opt *opt,
 			memcpy(&addr.s_addr, data, sizeof(addr.s_addr));
 			sl = snprintf(s, len, "%s", inet_ntoa(addr));
 			data += sizeof(addr.s_addr);
-		}
-#ifdef INET6
-		else if (opt->type & ADDRIPV6) {
-			ssize_t r;
-
-			r = ipv6_printaddr(s, len, data, ifname);
-			if (r != -1)
-				sl = r;
-			else
-				sl = 0;
+		} else if (opt->type & ADDRIPV6) {
+			sl = ipv6_printaddr(s, len, data, ifname);
 			data += 16;
+		} else {
+			errno = EINVAL;
+			return -1;
 		}
-#endif
-		else
-			sl = 0;
+		if (sl == -1)
+			return bytes == 0 ? -1 : bytes;
 		len -= (size_t)sl;
 		bytes += sl;
 		s += sl;
@@ -835,7 +850,7 @@ dhcp_set_leasefile(char *leasefile, size_t len, int family,
 }
 
 static size_t
-dhcp_envoption1(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
+dhcp_envoption1(char **env, const char *prefix,
     const struct dhcp_opt *opt, int vname, const uint8_t *od, size_t ol,
     const char *ifname)
 {
@@ -843,8 +858,11 @@ dhcp_envoption1(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
 	size_t e;
 	char *v, *val;
 
-	if (opt->len && opt->len < ol)
-		ol = opt->len;
+	/* Ensure a valid length */
+	ol = (size_t)dhcp_optlen(opt, ol);
+	if ((ssize_t)ol == -1)
+		return 0;
+
 	len = print_option(NULL, 0, opt, od, ol, ifname);
 	if (len < 0)
 		return 0;
@@ -858,10 +876,8 @@ dhcp_envoption1(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
 	if (env == NULL)
 		return e;
 	v = val = *env = malloc(e);
-	if (v == NULL) {
-		logger(ctx, LOG_ERR, "%s: %m", __func__);
+	if (v == NULL)
 		return 0;
-	}
 	if (vname)
 		v += snprintf(val, e, "%s_%s=", prefix, opt->var);
 	else
@@ -889,10 +905,14 @@ dhcp_envoption(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
 
 	/* If no embedded or encapsulated options, it's easy */
 	if (opt->embopts_len == 0 && opt->encopts_len == 0) {
-		if (!(opt->type & RESERVED) &&
-		    dhcp_envoption1(ctx, env == NULL ? NULL : &env[0],
-		    prefix, opt, 1, od, ol, ifname))
-			return 1;
+		if (!(opt->type & RESERVED)) {
+			if (dhcp_envoption1(env == NULL ? NULL : &env[0],
+			    prefix, opt, 1, od, ol, ifname))
+				return 1;
+			else
+				logger(ctx, LOG_ERR, "%s: %s %d: %m",
+				    ifname, __func__, opt->option);
+		}
 		return 0;
 	}
 
@@ -928,8 +948,8 @@ dhcp_envoption(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
 		if (eo == -1) {
 			if (env == NULL)
 				logger(ctx, LOG_ERR,
-				    "%s: %s: malformed embedded option"
-				    " %d:%d/%zu",
+				    "%s: %s %d.%d/%zu: "
+				    "malformed embedded option",
 				    ifname, __func__, opt->option,
 				    eopt->option, i);
 			goto out;
@@ -943,7 +963,7 @@ dhcp_envoption(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
 			if (env == NULL &&
 			    (ol != 0 || !(eopt->type & OPTIONAL)))
 				logger(ctx, LOG_ERR,
-				    "%s: %s: missing embedded option %d:%d/%zu",
+				    "%s: %s %d.%d/%zu: missing embedded option",
 				    ifname, __func__, opt->option,
 				    eopt->option, i);
 			goto out;
@@ -953,9 +973,14 @@ dhcp_envoption(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
 		 * This avoids new_fqdn_fqdn which would be silly. */
 		if (!(eopt->type & RESERVED)) {
 			ov = strcmp(opt->var, eopt->var);
-			if (dhcp_envoption1(ctx, env == NULL ? NULL : &env[n],
+			if (dhcp_envoption1(env == NULL ? NULL : &env[n],
 			    pfx, eopt, ov, od, (size_t)eo, ifname))
 				n++;
+			else if (env == NULL)
+				logger(ctx, LOG_ERR,
+				    "%s: %s %d.%d/%zu: %m",
+				    ifname, __func__,
+				    opt->option, eopt->option, i);
 		}
 		od += (size_t)eo;
 		ol -= (size_t)eo;
