@@ -1,6 +1,6 @@
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2015 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2016 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -96,7 +96,7 @@ static const struct dhcp_op dhcp_ops[] = {
 	{ DHCP_NAK,        "NAK" },
 	{ DHCP_RELEASE,    "RELEASE" },
 	{ DHCP_INFORM,     "INFORM" },
-	{ DHCP_FORCERENEW, "DHCP_FORCERENEW"},
+	{ DHCP_FORCERENEW, "FORCERENEW"},
 	{ 0, NULL }
 };
 
@@ -1224,12 +1224,13 @@ dhcp_getoption(struct dhcpcd_ctx *ctx,
 		*os = 2; /* code + len */
 		*code = (unsigned int)*od++;
 		*len = (size_t)*od++;
-		if (*len > ol) {
-			errno = EINVAL;
+		if (*len > ol - *os) {
+			errno = ERANGE;
 			return NULL;
 		}
 	}
 
+	*oopt = NULL;
 	for (i = 0, opt = ctx->dhcp_opts; i < ctx->dhcp_opts_len; i++, opt++) {
 		if (opt->option == *code) {
 			*oopt = opt;
@@ -1488,7 +1489,7 @@ dhcp_openudp(struct interface *ifp)
 	char *p;
 #endif
 
-	if ((s = xsocket(PF_INET, SOCK_DGRAM, IPPROTO_UDP, O_CLOEXEC)) == -1)
+	if ((s = xsocket(PF_INET, SOCK_DGRAM, IPPROTO_UDP, SOCK_CLOEXEC)) == -1)
 		return -1;
 
 	n = 1;
@@ -1931,8 +1932,10 @@ dhcp_arp_probed(struct arp_state *astate)
 		state->new = oldnew;
 	}
 #endif
-		
-	arp_announce(astate);
+
+	/* If we forked, stop here. */
+	if (astate->iface->ctx->options & DHCPCD_FORKED)
+		return;
 
 	/* Stop IPv4LL now we have a working DHCP address */
 	ipv4ll_drop(astate->iface);
@@ -2135,6 +2138,9 @@ dhcp_timeout(void *arg)
 	struct dhcp_state *state = D_STATE(ifp);
 
 	dhcp_bind(ifp);
+	/* If we forked, stop here. */
+	if (ifp->ctx->options & DHCPCD_FORKED)
+		return;
 	state->interval = 0;
 	dhcp_discover(ifp);
 }
@@ -2164,14 +2170,15 @@ dhcp_message_new(const struct in_addr *addr, const struct in_addr *mask)
 static int
 dhcp_arp_address(struct interface *ifp)
 {
-	const struct dhcp_state *state;
+	struct dhcp_state *state;
 	struct in_addr addr;
 	struct ipv4_addr *ia;
 	struct arp_state *astate;
 
 	eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
 
-	state = D_CSTATE(ifp);
+	state = D_STATE(ifp);
+	state->state = DHS_PROBE;
 	addr.s_addr = state->offer->yiaddr == INADDR_ANY ?
 	    state->offer->ciaddr : state->offer->yiaddr;
 	/* If the interface already has the address configured
@@ -2635,7 +2642,7 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 		/* Before we supported FORCERENEW we closed off the raw
 		 * port so we effectively ignored all messages.
 		 * As such we'll not log by default here. */
-		//log_dhcp(LOG_DEBUG, "bound, ignoring", iface, dhcp, from);
+		//log_dhcp(LOG_DEBUG, "bound, ignoring", ifp, dhcp, from);
 		return;
 	}
 
@@ -2647,6 +2654,13 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 		    inet_ntoa(*from));
 		return;
 	}
+
+	if (state->state == DHS_PROBE) {
+		/* Ignore any DHCP messages whilst probing a lease to bind. */
+		log_dhcp(LOG_DEBUG, "probing, ignoring", ifp, dhcp, from);
+		return;
+	}
+
 	/* reset the message counter */
 	state->interval = 0;
 
@@ -2783,6 +2797,16 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 		    get_option_addr(ifp->ctx,
 		    &lease->server, dhcp, DHO_SERVERID) != 0)
 			lease->server.s_addr = INADDR_ANY;
+
+		/* Test for rapid commit in the OFFER */
+		if (!(ifp->ctx->options & DHCPCD_TEST) &&
+		    has_option_mask(ifo->requestmask, DHO_RAPIDCOMMIT) &&
+		    get_option(ifp->ctx, dhcp, DHO_RAPIDCOMMIT, NULL))
+		{
+			state->state = DHS_REQUEST;
+			goto rapidcommit;
+		}
+
 		log_dhcp(LOG_INFO, "offered", ifp, dhcp, from);
 		free(state->offer);
 		state->offer = dhcp;
@@ -2825,6 +2849,20 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 			return;
 		}
 
+		if (state->state == DHS_DISCOVER) {
+			/* We only allow ACK of rapid commit DISCOVER. */
+			if (has_option_mask(ifo->requestmask,
+			    DHO_RAPIDCOMMIT) &&
+			    get_option(ifp->ctx, dhcp, DHO_RAPIDCOMMIT, NULL))
+				state->state = DHS_REQUEST;
+			else {
+				log_dhcp(LOG_DEBUG, "ignoring ack of",
+				    ifp, dhcp, from);
+				return;
+			}
+		}
+
+rapidcommit:
 		if (!(ifo->options & DHCPCD_INFORM))
 			log_dhcp(LOG_DEBUG, "acknowledged", ifp, dhcp, from);
 		else
@@ -3051,7 +3089,7 @@ dhcp_open(struct interface *ifp)
 			return -1;
 		}
 		eloop_event_add(ifp->ctx->eloop,
-		    state->raw_fd, dhcp_handlepacket, ifp, NULL, NULL);
+		    state->raw_fd, dhcp_handlepacket, ifp);
 	}
 	return 0;
 }
@@ -3229,8 +3267,7 @@ dhcp_start1(void *arg)
 				    "%s: dhcp_openudp: %m", __func__);
 		} else
 			eloop_event_add(ifp->ctx->eloop,
-			    ifp->ctx->udp_fd, dhcp_handleudp,
-			    ifp->ctx, NULL, NULL);
+			    ifp->ctx->udp_fd, dhcp_handleudp, ifp->ctx);
 	}
 
 	if (dhcp_init(ifp) == -1) {
