@@ -1,6 +1,6 @@
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2015 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2016 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -35,13 +35,12 @@
 #include <sys/uio.h>
 #include <sys/utsname.h>
 
+#include "config.h"
+
 #include <arpa/inet.h>
 #include <net/bpf.h>
 #include <net/if.h>
 #include <net/if_dl.h>
-#ifdef __FreeBSD__ /* Needed so that including netinet6/in6_var.h works */
-#  include <net/if_var.h>
-#endif
 #include <net/if_media.h>
 #include <net/route.h>
 #include <netinet/if_ether.h>
@@ -75,7 +74,6 @@
 #undef IPV6CTL_ACCEPT_RTADV
 #endif
 
-#include "config.h"
 #include "common.h"
 #include "dhcp.h"
 #include "if.h"
@@ -95,18 +93,22 @@
 
 #define COPYOUT(sin, sa) do {						      \
 	if ((sa) && ((sa)->sa_family == AF_INET || (sa)->sa_family == 255))   \
-		(sin) = ((struct sockaddr_in*)(void *)(sa))->sin_addr;	      \
+		(sin) = ((struct sockaddr_in *)(void *)(sa))->sin_addr;	      \
 	} while (0)
 
 #define COPYOUT6(sin, sa) do {						      \
 	if ((sa) && ((sa)->sa_family == AF_INET6 || (sa)->sa_family == 255))  \
-		(sin) = ((struct sockaddr_in6*)(void *)(sa))->sin6_addr;      \
+		(sin) = ((struct sockaddr_in6 *)(void *)(sa))->sin6_addr;     \
 	} while (0)
 
 #ifndef CLLADDR
 #  define CLLADDR(s) ((const char *)((s)->sdl_data + (s)->sdl_nlen))
 #endif
 
+struct priv {
+	int pf_inet6_fd;
+};
+
 int
 if_init(__unused struct interface *iface)
 {
@@ -122,10 +124,36 @@ if_conf(__unused struct interface *iface)
 }
 
 int
-if_openlinksocket(void)
+if_opensockets_os(struct dhcpcd_ctx *ctx)
 {
+	struct priv *priv;
 
-	return xsocket(PF_ROUTE, SOCK_RAW, 0, O_NONBLOCK|O_CLOEXEC);
+	if ((priv = malloc(sizeof(*priv))) == NULL)
+		return -1;
+	ctx->priv = priv;
+
+#ifdef INET6
+	priv->pf_inet6_fd = xsocket(PF_INET6, SOCK_DGRAM, 0, SOCK_CLOEXEC);
+	/* Don't return an error so we at least work on kernels witout INET6
+	 * even though we expect INET6 support.
+	 * We will fail noisily elsewhere anyway. */
+#else
+	priv->pf_inet6_fd = -1;
+#endif
+
+	ctx->link_fd = xsocket(PF_ROUTE, SOCK_RAW, 0,
+	    SOCK_CLOEXEC | SOCK_NONBLOCK);
+	return ctx->link_fd == -1 ? -1 : 0;
+}
+
+void
+if_closesockets_os(struct dhcpcd_ctx *ctx)
+{
+	struct priv *priv;
+
+	priv = (struct priv *)ctx->priv;
+	if (priv->pf_inet6_fd != -1)
+		close(priv->pf_inet6_fd);
 }
 
 #if defined(INET) || defined(INET6)
@@ -150,7 +178,7 @@ if_getssid1(int s, const char *ifname, uint8_t *ssid)
 	struct ieee80211_nwid nwid;
 #elif defined(IEEE80211_IOC_SSID)
 	struct ieee80211req ireq;
-	char nwid[IEEE80211_NWID_LEN + 1];
+	char nwid[IEEE80211_NWID_LEN];
 #endif
 
 #if defined(SIOCG80211NWID) /* NetBSD */
@@ -161,13 +189,11 @@ if_getssid1(int s, const char *ifname, uint8_t *ssid)
 	if (ioctl(s, SIOCG80211NWID, &ifr) == 0) {
 		if (ssid == NULL)
 			retval = nwid.i_len;
-		else if (nwid.i_len > IF_SSIDSIZE) {
+		else if (nwid.i_len > IF_SSIDLEN)
 			errno = ENOBUFS;
-			retval = -1;
-		} else {
+		else {
 			retval = nwid.i_len;
 			memcpy(ssid, nwid.i_nwid, nwid.i_len);
-			ssid[nwid.i_len] = '\0';
 		}
 	}
 #elif defined(IEEE80211_IOC_SSID) /* FreeBSD */
@@ -180,15 +206,15 @@ if_getssid1(int s, const char *ifname, uint8_t *ssid)
 	if (ioctl(s, SIOCG80211, &ireq) == 0) {
 		if (ssid == NULL)
 			retval = ireq.i_len;
-		else if (ireq.i_len > IF_SSIDSIZE) {
+		else if (ireq.i_len > IF_SSIDLEN)
 			errno = ENOBUFS;
-			retval = -1;
-		} else  {
+		else  {
 			retval = ireq.i_len;
 			memcpy(ssid, nwid, ireq.i_len);
-			ssid[ireq.i_len] = '\0';
 		}
 	}
+#else
+	errno = ENOSYS;
 #endif
 
 	return retval;
@@ -202,6 +228,9 @@ if_getssid(struct interface *ifp)
 	r = if_getssid1(ifp->ctx->pf_inet_fd, ifp->name, ifp->ssid);
 	if (r != -1)
 		ifp->ssid_len = (unsigned int)r;
+	else
+		ifp->ssid_len = 0;
+	ifp->ssid[ifp->ssid_len] = '\0';
 	return r;
 }
 
@@ -251,12 +280,27 @@ static struct interface *
 if_findsdl(struct dhcpcd_ctx *ctx, struct sockaddr_dl *sdl)
 {
 
+	if (sdl->sdl_index)
+		return if_findindex(ctx->ifaces, sdl->sdl_index);
+
 	if (sdl->sdl_nlen) {
 		char ifname[IF_NAMESIZE];
+
 		memcpy(ifname, sdl->sdl_data, sdl->sdl_nlen);
 		ifname[sdl->sdl_nlen] = '\0';
 		return if_find(ctx->ifaces, ifname);
 	}
+	if (sdl->sdl_alen) {
+		struct interface *ifp;
+
+		TAILQ_FOREACH(ifp, ctx->ifaces, next) {
+			if (ifp->hwlen == sdl->sdl_alen &&
+			    memcmp(ifp->hwaddr,
+			    sdl->sdl_data, sdl->sdl_alen) == 0)
+				return ifp;
+		}
+	}
+
 	return NULL;
 }
 #endif
@@ -277,21 +321,40 @@ if_openrawsocket(struct interface *ifp, uint16_t protocol)
 #ifdef BIOCIMMEDIATE
 	int flags;
 #endif
+#ifndef O_CLOEXEC
+	int fd_opts;
+#endif
+
 #ifdef _PATH_BPF
-	fd = open(_PATH_BPF, O_RDWR | O_CLOEXEC | O_NONBLOCK);
+	fd = open(_PATH_BPF, O_RDWR | O_NONBLOCK
+#ifdef O_CLOEXEC
+		| O_CLOEXEC
+#endif
+	);
 #else
 	char device[32];
 	int n = 0;
 
 	do {
 		snprintf(device, sizeof(device), "/dev/bpf%d", n++);
-		fd = open(device, O_RDWR | O_CLOEXEC | O_NONBLOCK);
+		fd = open(device, O_RDWR | O_NONBLOCK
+#ifdef O_CLOEXEC
+				| O_CLOEXEC
+#endif
+		);
 	} while (fd == -1 && errno == EBUSY);
 #endif
 
 	if (fd == -1)
 		return -1;
 
+#ifndef O_CLOEXEC
+	if ((fd_opts = fcntl(fd, F_GETFD)) == -1 ||
+	    fcntl(fd, F_SETFD, fd_opts | FD_CLOEXEC) == -1) {
+		close(fd);
+		return -1;
+	}
+#endif
 	state = IPV4_STATE(ifp);
 	memset(&pv, 0, sizeof(pv));
 	if (ioctl(fd, BIOCVERSION, &pv) == -1)
@@ -448,9 +511,12 @@ if_copyrt(struct dhcpcd_ctx *ctx, struct rt *rt, struct rt_msghdr *rtm)
 {
 	char *cp;
 	struct sockaddr *sa, *rti_info[RTAX_MAX];
+	struct sockaddr_dl *sdl;
+	struct sockaddr_in *sin;
+	struct ipv4_addr *ia;
 
-	cp = (char *)(void *)(rtm + 1);
-	sa = (struct sockaddr *)(void *)cp;
+	cp = (void *)(rtm + 1);
+	sa = (void *)cp;
 	if (sa->sa_family != AF_INET)
 		return -1;
 	if (~rtm->rtm_addrs & (RTA_DST | RTA_GATEWAY))
@@ -478,17 +544,30 @@ if_copyrt(struct dhcpcd_ctx *ctx, struct rt *rt, struct rt_msghdr *rtm)
 		rt->net.s_addr = INADDR_BROADCAST;
 	COPYOUT(rt->gate, rti_info[RTAX_GATEWAY]);
 	COPYOUT(rt->src, rti_info[RTAX_IFA]);
-
-	if (rtm->rtm_inits & RTV_MTU)
-		rt->mtu = (unsigned int)rtm->rtm_rmx.rmx_mtu;
+	rt->mtu = (unsigned int)rtm->rtm_rmx.rmx_mtu;
 
 	if (rtm->rtm_index)
 		rt->iface = if_findindex(ctx->ifaces, rtm->rtm_index);
 	else if (rtm->rtm_addrs & RTA_IFP) {
-		struct sockaddr_dl *sdl;
-
-		sdl = (struct sockaddr_dl *)(void *)rti_info[RTAX_IFP];
+		sdl = (void *)rti_info[RTAX_IFP];
 		rt->iface = if_findsdl(ctx, sdl);
+	} else if (rtm->rtm_addrs & RTA_GATEWAY) {
+		sa = rti_info[RTAX_GATEWAY];
+		switch (sa->sa_family) {
+		case AF_LINK:
+			sdl = (void *)sa;
+			rt->iface = if_findsdl(ctx, sdl);
+			break;
+		case AF_INET:
+			sin = (void *)sa;
+			if ((ia = ipv4_findmaskaddr(ctx, &sin->sin_addr)))
+				rt->iface = ia->iface;
+			break;
+		default:
+			errno = EAFNOSUPPORT;
+			logger(ctx, LOG_ERR, "%s: %m", __func__);
+			return -1;
+		}
 	}
 
 	/* If we don't have an interface and it's a host route, it maybe
@@ -496,12 +575,14 @@ if_copyrt(struct dhcpcd_ctx *ctx, struct rt *rt, struct rt_msghdr *rtm)
 	if (rt->iface == NULL &&
 	    !(~rtm->rtm_flags & (RTF_HOST | RTF_GATEWAY)))
 	{
-		struct ipv4_addr *ia;
-
 		if ((ia = ipv4_findaddr(ctx, &rt->dest)))
 			rt->iface = ia->iface;
 	}
 
+	if (rt->iface == NULL) {
+		errno = ESRCH;
+		return -1;
+	}
 	return 0;
 }
 
@@ -522,6 +603,11 @@ if_route(unsigned char cmd, const struct rt *rt)
 	size_t l;
 	struct in_addr src_addr;
 
+	if ((cmd == RTM_ADD || cmd == RTM_DELETE || cmd == RTM_CHANGE) &&
+	    rt->iface->ctx->options & DHCPCD_DAEMONISE &&
+	    !(rt->iface->ctx->options & DHCPCD_DAEMONISED))
+		rt->iface->ctx->options |= DHCPCD_RTM_PPID;
+
 #define ADDSU {								      \
 		l = RT_ROUNDUP(su.sa.sa_len);				      \
 		memcpy(bp, &su, l);					      \
@@ -537,11 +623,11 @@ if_route(unsigned char cmd, const struct rt *rt)
 
 	memset(&rtm, 0, sizeof(rtm));
 	rtm.hdr.rtm_version = RTM_VERSION;
-	rtm.hdr.rtm_seq = 1;
 	rtm.hdr.rtm_type = cmd;
 	rtm.hdr.rtm_addrs = RTA_DST;
 	rtm.hdr.rtm_flags = RTF_UP;
 	rtm.hdr.rtm_pid = getpid();
+	rtm.hdr.rtm_seq = ++rt->iface->ctx->seq;
 #ifdef RTF_PINNED
 	if (cmd != RTM_ADD)
 		rtm.hdr.rtm_flags |= RTF_PINNED;
@@ -558,6 +644,9 @@ if_route(unsigned char cmd, const struct rt *rt)
 #ifdef RTF_CLONING
 			rtm.hdr.rtm_flags |= RTF_CLONING;
 #endif
+#ifdef RTF_CONNECTED
+			rtm.hdr.rtm_flags |= RTF_CONNECTED;
+#endif
 #ifdef RTP_CONNECTED
 			rtm.hdr.rtm_priority = RTP_CONNECTED;
 #endif
@@ -605,6 +694,9 @@ if_route(unsigned char cmd, const struct rt *rt)
 		    rt->gate.s_addr == htonl(INADDR_ANY)) ||
 #ifdef RTF_CLONING
 		    rtm.hdr.rtm_flags & RTF_CLONING ||
+#endif
+#ifdef RTF_CONNECTED
+		    rtm.hdr.rtm_flags & RTF_CONNECTED ||
 #endif
 		    !(rtm.hdr.rtm_flags & RTF_STATIC))
 		{
@@ -639,8 +731,10 @@ if_route(unsigned char cmd, const struct rt *rt)
 #undef ADDSU
 
 	rtm.hdr.rtm_msglen = (unsigned short)(bp - (char *)&rtm);
-	return write(rt->iface->ctx->link_fd,
-	    &rtm, rtm.hdr.rtm_msglen) == -1 ? -1 : 0;
+	if (write(rt->iface->ctx->link_fd, &rtm, rtm.hdr.rtm_msglen) == -1)
+		return -1;
+	rt->iface->ctx->sseq = rt->iface->ctx->seq;
+	return 0;
 }
 
 int
@@ -672,9 +766,9 @@ if_initrt(struct interface *ifp)
 
 	end = buf + needed;
 	for (p = buf; p < end; p += rtm->rtm_msglen) {
-		rtm = (struct rt_msghdr *)(void *)p;
+		rtm = (void *)p;
 		if (if_copyrt(ifp->ctx, &rt, rtm) == 0)
-			ipv4_handlert(ifp->ctx, RTM_ADD, &rt);
+			ipv4_handlert(ifp->ctx, RTM_ADD, &rt, 1);
 	}
 	free(buf);
 	return 0;
@@ -689,7 +783,7 @@ if_addrflags(const struct in_addr *addr, const struct interface *ifp)
 
 	memset(&ifr, 0, sizeof(ifr));
 	strlcpy(ifr.ifr_name, ifp->name, sizeof(ifr.ifr_name));
-	sin = (struct sockaddr_in *)(void *)&ifr.ifr_addr;
+	sin = (void *)&ifr.ifr_addr;
 	sin->sin_family = AF_INET;
 	sin->sin_addr = *addr;
 	if (ioctl(ifp->ctx->pf_inet_fd, SIOCGIFAFLAG_IN, &ifr) == -1)
@@ -744,6 +838,7 @@ if_address6(const struct ipv6_addr *ia, int action)
 {
 	struct in6_aliasreq ifa;
 	struct in6_addr mask;
+	struct priv *priv;
 
 	memset(&ifa, 0, sizeof(ifa));
 	strlcpy(ifa.ifra_name, ia->iface->name, sizeof(ifa.ifra_name));
@@ -778,7 +873,8 @@ if_address6(const struct ipv6_addr *ia, int action)
 	ifa.ifra_lifetime.ia6t_pltime = ia->prefix_pltime;
 #undef ADDADDR
 
-	return ioctl(ia->iface->ctx->pf_inet6_fd,
+	priv = (struct priv *)ia->iface->ctx->priv;
+	return ioctl(priv->pf_inet6_fd,
 	    action < 0 ? SIOCDIFADDR_IN6 : SIOCAIFADDR_IN6, &ifa);
 }
 
@@ -788,9 +884,12 @@ if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct rt_msghdr *rtm)
 {
 	char *cp;
 	struct sockaddr *sa, *rti_info[RTAX_MAX];
+	struct sockaddr_dl *sdl;
+	struct sockaddr_in6 *sin;
+	struct ipv6_addr *ia;
 
-	cp = (char *)(void *)(rtm + 1);
-	sa = (struct sockaddr *)(void *)cp;
+	cp = (void *)(rtm + 1);
+	sa = (void *)cp;
 	if (sa->sa_family != AF_INET6)
 		return -1;
 	if (~rtm->rtm_addrs & (RTA_DST | RTA_GATEWAY))
@@ -811,6 +910,7 @@ if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct rt_msghdr *rtm)
 	memset(rt, 0, sizeof(*rt));
 	rt->flags = (unsigned int)rtm->rtm_flags;
 	COPYOUT6(rt->dest, rti_info[RTAX_DST]);
+	DESCOPE(&rt->dest);
 	if (rtm->rtm_addrs & RTA_NETMASK) {
 		/*
 		 * We need to zero out the struct beyond sin6_len and
@@ -820,19 +920,25 @@ if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct rt_msghdr *rtm)
 		 * Either way it needs to be zeroed out.
 		 */
 		struct sockaddr_in6 *sin6;
-		size_t e, i, len = 0, final = 0;
+		size_t e, i, final = 0, illegal = 0;
+		const unsigned char *p;
 
-		sin6 = (struct sockaddr_in6 *)(void *)rti_info[RTAX_NETMASK];
+		sin6 = (void *)rti_info[RTAX_NETMASK];
 		rt->net = sin6->sin6_addr;
 		e = sin6->sin6_len - offsetof(struct sockaddr_in6, sin6_addr);
 		if (e > sizeof(struct in6_addr))
 			e = sizeof(struct in6_addr);
-		for (i = 0; i < e; i++) {
-			switch (rt->net.s6_addr[i] & 0xff) {
+		for (p = (const unsigned char *)&sin6->sin6_addr, i = 0;
+		    i < e;
+		    p++)
+		{
+			if (final && *p) {
+				illegal = 1;
+				rt->net.s6_addr[i++] = 0x00;
+				continue;
+			}
+			switch (*p & 0xff) {
 			case 0xff:
-				/* We don't really want the length,
-				 * just that it's valid */
-				len++;
 				break;
 			case 0xfe:
 			case 0xfc:
@@ -841,47 +947,63 @@ if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct rt_msghdr *rtm)
 			case 0xe0:
 			case 0xc0:
 			case 0x80:
-				len++;
 				final = 1;
 				break;
 			default:
-				rt->net.s6_addr[i] = 0x00;
 				final = 1;
+				illegal = 1;
 				break;
 			}
-			if (final)
-				break;
+			if (!illegal)
+				rt->net.s6_addr[i++] &= *p;
+			else
+				rt->net.s6_addr[i++] = 0x00;
 		}
-		if (len == 0)
-			i = 0;
 		while (i < sizeof(rt->net.s6_addr))
 			rt->net.s6_addr[i++] = 0x00;
 	} else
 		ipv6_mask(&rt->net, 128);
 	COPYOUT6(rt->gate, rti_info[RTAX_GATEWAY]);
-
-	if (rtm->rtm_inits & RTV_MTU)
-		rt->mtu = (unsigned int)rtm->rtm_rmx.rmx_mtu;
+	DESCOPE(&rt->gate);
+	rt->mtu = (unsigned int)rtm->rtm_rmx.rmx_mtu;
 
 	if (rtm->rtm_index)
 		rt->iface = if_findindex(ctx->ifaces, rtm->rtm_index);
 	else if (rtm->rtm_addrs & RTA_IFP) {
-		struct sockaddr_dl *sdl;
-
-		sdl = (struct sockaddr_dl *)(void *)rti_info[RTAX_IFP];
+		sdl = (void *)rti_info[RTAX_IFP];
 		rt->iface = if_findsdl(ctx, sdl);
+	} else if (rtm->rtm_addrs & RTA_GATEWAY) {
+		sa = rti_info[RTAX_GATEWAY];
+		switch (sa->sa_family) {
+		case AF_LINK:
+			sdl = (void *)sa;
+			rt->iface = if_findsdl(ctx, sdl);
+			break;
+		case AF_INET6:
+			sin = (void *)sa;
+			if ((ia = ipv6_findmaskaddr(ctx, &sin->sin6_addr)))
+				rt->iface = ia->iface;
+			break;
+		default:
+			errno = EAFNOSUPPORT;
+			logger(ctx, LOG_ERR, "%s: %m", __func__);
+			return -1;
+		}
 	}
+
 	/* If we don't have an interface and it's a host route, it maybe
 	 * to a local ip via the loopback interface. */
 	if (rt->iface == NULL &&
 	    !(~rtm->rtm_flags & (RTF_HOST | RTF_GATEWAY)))
 	{
-		struct ipv6_addr *ia;
-
 		if ((ia = ipv6_findaddr(ctx, &rt->dest, 0)))
 			rt->iface = ia->iface;
 	}
 
+	if (rt->iface == NULL) {
+		errno = ESRCH;
+		return -1;
+	}
 	return 0;
 }
 
@@ -901,6 +1023,11 @@ if_route6(unsigned char cmd, const struct rt6 *rt)
 	char *bp = rtm.buffer;
 	size_t l;
 
+	if ((cmd == RTM_ADD || cmd == RTM_DELETE || cmd == RTM_CHANGE) &&
+	    rt->iface->ctx->options & DHCPCD_DAEMONISE &&
+	    !(rt->iface->ctx->options & DHCPCD_DAEMONISED))
+		rt->iface->ctx->options |= DHCPCD_RTM_PPID;
+
 #define ADDSU {								      \
 		l = RT_ROUNDUP(su.sa.sa_len);				      \
 		memcpy(bp, &su, l);					      \
@@ -919,10 +1046,10 @@ if_route6(unsigned char cmd, const struct rt6 *rt)
 
 	memset(&rtm, 0, sizeof(rtm));
 	rtm.hdr.rtm_version = RTM_VERSION;
-	rtm.hdr.rtm_seq = 1;
 	rtm.hdr.rtm_type = cmd;
 	rtm.hdr.rtm_flags = RTF_UP | (int)rt->flags;
 	rtm.hdr.rtm_pid = getpid();
+	rtm.hdr.rtm_seq = ++rt->iface->ctx->seq;
 #ifdef RTF_PINNED
 	if (rtm.hdr.rtm_type != RTM_ADD)
 		rtm.hdr.rtm_flags |= RTF_PINNED;
@@ -933,16 +1060,20 @@ if_route6(unsigned char cmd, const struct rt6 *rt)
 #ifdef RTF_CLONING
 		rtm.hdr.rtm_flags |= RTF_CLONING;
 #endif
+#ifdef RTF_CONNECTED
+		rtm.hdr.rtm_flags |= RTF_CONNECTED;
+#endif
 #ifdef RTP_CONNECTED
 		rtm.hdr.rtm_priority = RTP_CONNECTED;
 #endif
 	} else
 		rtm.hdr.rtm_flags |= RTF_GATEWAY | RTF_STATIC;
 
-	if (cmd == RTM_ADD)
+	if (cmd == RTM_ADD || cmd == RTM_CHANGE) {
 		rtm.hdr.rtm_addrs |= RTA_GATEWAY;
-	if (cmd == RTM_ADD && !(rtm.hdr.rtm_flags & RTF_REJECT))
-		rtm.hdr.rtm_addrs |= RTA_IFP | RTA_IFA;
+	    	if (!(rtm.hdr.rtm_flags & RTF_REJECT))
+			rtm.hdr.rtm_addrs |= RTA_IFP | RTA_IFA;
+	}
 
 	ADDADDR(&rt->dest);
 	if (rtm.hdr.rtm_addrs & RTA_GATEWAY) {
@@ -957,9 +1088,7 @@ if_route6(unsigned char cmd, const struct rt6 *rt)
 	if (rtm.hdr.rtm_addrs & RTA_NETMASK)
 		ADDADDR(&rt->net);
 
-	if ((cmd == RTM_ADD || cmd == RTM_CHANGE) &&
-	    (rtm.hdr.rtm_addrs & (RTA_IFP | RTA_IFA)))
-	{
+	if (rtm.hdr.rtm_addrs & (RTA_IFP | RTA_IFA)) {
 		rtm.hdr.rtm_index = (unsigned short)rt->iface->index;
 		if (rtm.hdr.rtm_addrs & RTA_IFP) {
 			if_linkaddr(&su.sdl, rt->iface);
@@ -969,7 +1098,7 @@ if_route6(unsigned char cmd, const struct rt6 *rt)
 		if (rtm.hdr.rtm_addrs & RTA_IFA) {
 			const struct ipv6_addr *lla;
 
-			lla = ipv6_linklocal(rt->iface);
+			lla = ipv6_linklocal(UNCONST(rt->iface));
 			if (lla == NULL) /* unlikely */
 					return -1;
 			ADDADDRS(&lla->addr, rt->iface->index);
@@ -985,8 +1114,10 @@ if_route6(unsigned char cmd, const struct rt6 *rt)
 #undef ADDSU
 
 	rtm.hdr.rtm_msglen = (unsigned short)(bp - (char *)&rtm);
-	return write(rt->iface->ctx->link_fd,
-	    &rtm, rtm.hdr.rtm_msglen) == -1 ? -1 : 0;
+	if (write(rt->iface->ctx->link_fd, &rtm, rtm.hdr.rtm_msglen) == -1)
+		return -1;
+	rt->iface->ctx->sseq = rt->iface->ctx->seq;
+	return 0;
 }
 
 int
@@ -1018,7 +1149,7 @@ if_initrt6(struct interface *ifp)
 
 	end = buf + needed;
 	for (p = buf; p < end; p += rtm->rtm_msglen) {
-		rtm = (struct rt_msghdr *)(void *)p;
+		rtm = (void *)p;
 		if (if_copyrt6(ifp->ctx, &rt, rtm) == 0)
 			ipv6_handlert(ifp->ctx, RTM_ADD, &rt);
 	}
@@ -1031,13 +1162,15 @@ if_addrflags6(const struct in6_addr *addr, const struct interface *ifp)
 {
 	int flags;
 	struct in6_ifreq ifr6;
+	struct priv *priv;
 
 	memset(&ifr6, 0, sizeof(ifr6));
 	strlcpy(ifr6.ifr_name, ifp->name, sizeof(ifr6.ifr_name));
 	ifr6.ifr_addr.sin6_family = AF_INET6;
 	ifr6.ifr_addr.sin6_addr = *addr;
 	ifa_scope(&ifr6.ifr_addr, ifp->index);
-	if (ioctl(ifp->ctx->pf_inet6_fd, SIOCGIFAFLAG_IN6, &ifr6) != -1)
+	priv = (struct priv *)ifp->ctx->priv;
+	if (ioctl(priv->pf_inet6_fd, SIOCGIFAFLAG_IN6, &ifr6) != -1)
 		flags = ifr6.ifr_ifru.ifru_flags6;
 	else
 		flags = -1;
@@ -1050,14 +1183,15 @@ if_getlifetime6(struct ipv6_addr *ia)
 	struct in6_ifreq ifr6;
 	time_t t;
 	struct in6_addrlifetime *lifetime;
+	struct priv *priv;
 
 	memset(&ifr6, 0, sizeof(ifr6));
 	strlcpy(ifr6.ifr_name, ia->iface->name, sizeof(ifr6.ifr_name));
 	ifr6.ifr_addr.sin6_family = AF_INET6;
 	ifr6.ifr_addr.sin6_addr = ia->addr;
 	ifa_scope(&ifr6.ifr_addr, ia->iface->index);
-	if (ioctl(ia->iface->ctx->pf_inet6_fd,
-	    SIOCGIFALIFETIME_IN6, &ifr6) == -1)
+	priv = (struct priv *)ia->iface->ctx->priv;
+	if (ioctl(priv->pf_inet6_fd, SIOCGIFALIFETIME_IN6, &ifr6) == -1)
 		return -1;
 
 	t = time(NULL);
@@ -1110,14 +1244,31 @@ if_managelink(struct dhcpcd_ctx *ctx)
 		return -1;
 	e = msg + bytes;
 	for (p = msg; p < e; p += rtm->rtm_msglen) {
-		rtm = (struct rt_msghdr *)(void *)p;
-		// Ignore messages generated by us
-		if (rtm->rtm_pid == getpid())
-			break;
+		rtm = (void *)p;
+		if (rtm->rtm_type == RTM_MISS)
+			continue;
+		/* Ignore messages generated by us */
+		if (rtm->rtm_pid == getpid()) {
+			ctx->options &= ~DHCPCD_RTM_PPID;
+			continue;
+		}
+		/* Ignore messages sent by the parent process after forking */
+		if ((ctx->options & (DHCPCD_RTM_PPID | DHCPCD_DAEMONISED)) ==
+		    (DHCPCD_RTM_PPID | DHCPCD_DAEMONISED) &&
+		    rtm->rtm_pid == ctx->ppid)
+		{
+			/* If this is the last successful message sent clear
+			 * the check flag as it's possible another process could
+			 * re-use the same pid and also manipulate the kernel
+			 * routing table. */
+			if (rtm->rtm_seq == ctx->pseq)
+				ctx->options &= ~DHCPCD_RTM_PPID;
+			continue;
+		}
 		switch(rtm->rtm_type) {
 #ifdef RTM_IFANNOUNCE
 		case RTM_IFANNOUNCE:
-			ifan = (struct if_announcemsghdr *)(void *)p;
+			ifan = (void *)p;
 			switch(ifan->ifan_what) {
 			case IFAN_ARRIVAL:
 				dhcpcd_handleinterface(ctx, 1,
@@ -1131,7 +1282,7 @@ if_managelink(struct dhcpcd_ctx *ctx)
 			break;
 #endif
 		case RTM_IFINFO:
-			ifm = (struct if_msghdr *)(void *)p;
+			ifm = (void *)p;
 			ifp = if_findindex(ctx->ifaces, ifm->ifm_index);
 			if (ifp == NULL)
 				break;
@@ -1159,13 +1310,13 @@ if_managelink(struct dhcpcd_ctx *ctx)
 		case RTM_ADD:
 		case RTM_CHANGE:
 		case RTM_DELETE:
-			cp = (char *)(void *)(rtm + 1);
-			sa = (struct sockaddr *)(void *)cp;
+			cp = (void *)(rtm + 1);
+			sa = (void *)cp;
 			switch (sa->sa_family) {
 #ifdef INET
 			case AF_INET:
 				if (if_copyrt(ctx, &rt, rtm) == 0)
-					ipv4_handlert(ctx, rtm->rtm_type, &rt);
+					ipv4_handlert(ctx, rtm->rtm_type,&rt,0);
 				break;
 #endif
 #ifdef INET6
@@ -1208,11 +1359,11 @@ if_managelink(struct dhcpcd_ctx *ctx)
 #endif
 		case RTM_DELADDR:	/* FALLTHROUGH */
 		case RTM_NEWADDR:
-			ifam = (struct ifa_msghdr *)(void *)p;
+			ifam = (void *)p;
 			ifp = if_findindex(ctx->ifaces, ifam->ifam_index);
 			if (ifp == NULL)
 				break;
-			cp = (char *)(void *)(ifam + 1);
+			cp = (void *)(ifam + 1);
 			get_addrs(ifam->ifam_addrs, cp, rti_info);
 			if (rti_info[RTAX_IFA] == NULL)
 				break;
@@ -1250,12 +1401,10 @@ if_managelink(struct dhcpcd_ctx *ctx)
 #endif
 #ifdef INET6
 			case AF_INET6:
-				sin6 = (struct sockaddr_in6*)(void *)
-				    rti_info[RTAX_IFA];
+				sin6 = (void *)rti_info[RTAX_IFA];
 				ia6 = sin6->sin6_addr;
 				DESCOPE(&ia6);
-				sin6 = (struct sockaddr_in6*)(void *)
-				    rti_info[RTAX_NETMASK];
+				sin6 = (void *)rti_info[RTAX_NETMASK];
 				net6 = sin6->sin6_addr;
 				DESCOPE(&net6);
 				if (rtm->rtm_type == RTM_NEWADDR) {
@@ -1623,7 +1772,9 @@ _if_checkipv6(int s, struct dhcpcd_ctx *ctx,
 int
 if_checkipv6(struct dhcpcd_ctx *ctx, const struct interface *ifp, int own)
 {
+	struct priv *priv;
 
-	return _if_checkipv6(ctx->pf_inet6_fd, ctx, ifp, own);
+	priv = (struct priv *)ctx->priv;
+	return _if_checkipv6(priv->pf_inet6_fd, ctx, ifp, own);
 }
 #endif
