@@ -1,6 +1,6 @@
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2015 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2016 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -91,6 +91,11 @@ int if_getssid_wext(const char *ifname, uint8_t *ssid);
 
 #include "bpf-filter.h"
 
+struct priv {
+	int route_fd;
+	uint32_t route_pid;
+};
+
 /* Broadcast address for IPoIB */
 static const uint8_t ipv4_bcast_addr[] = {
 	0x00, 0xff, 0xff, 0xff,
@@ -248,7 +253,7 @@ _open_link_socket(struct sockaddr_nl *nl, int protocol)
 {
 	int fd;
 
-	if ((fd = xsocket(AF_NETLINK, SOCK_RAW, protocol, O_CLOEXEC)) == -1)
+	if ((fd = xsocket(AF_NETLINK, SOCK_RAW, protocol, SOCK_CLOEXEC)) == -1)
 		return -1;
 	nl->nl_family = AF_NETLINK;
 	if (bind(fd, (struct sockaddr *)nl, sizeof(*nl)) == -1) {
@@ -259,10 +264,15 @@ _open_link_socket(struct sockaddr_nl *nl, int protocol)
 }
 
 int
-if_openlinksocket(void)
+if_opensockets_os(struct dhcpcd_ctx *ctx)
 {
+	struct priv *priv;
 	struct sockaddr_nl snl;
+	socklen_t len;
 
+	/* Open the link socket first so it gets pid() for the socket.
+	 * Then open our persistent route socket so we get a unique
+	 * pid that doesn't clash with a process id for after we fork. */
 	memset(&snl, 0, sizeof(snl));
 	snl.nl_groups = RTMGRP_LINK;
 
@@ -273,7 +283,34 @@ if_openlinksocket(void)
 	snl.nl_groups |= RTMGRP_IPV6_ROUTE | RTMGRP_IPV6_IFADDR | RTMGRP_NEIGH;
 #endif
 
-	return _open_link_socket(&snl, NETLINK_ROUTE);
+	ctx->link_fd = _open_link_socket(&snl, NETLINK_ROUTE);
+	if (ctx->link_fd == -1)
+		return -1;
+
+	if ((priv = calloc(1, sizeof(*priv))) == NULL)
+		return -1;
+
+	ctx->priv = priv;
+	memset(&snl, 0, sizeof(snl));
+	priv->route_fd = _open_link_socket(&snl, NETLINK_ROUTE);
+	if (priv->route_fd == -1)
+		return -1;
+	len = sizeof(snl);
+	if (getsockname(priv->route_fd, (struct sockaddr *)&snl, &len) == -1)
+		return -1;
+	priv->route_pid = snl.nl_pid;
+	return 0;
+}
+
+void
+if_closesockets_os(struct dhcpcd_ctx *ctx)
+{
+	struct priv *priv;
+
+	if (ctx->priv != NULL) {
+		priv = (struct priv *)ctx->priv;
+		close(priv->route_fd);
+	}
 }
 
 static int
@@ -330,7 +367,7 @@ get_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp, int fd, int flags,
 				goto eexit;
 			buf = nbuf;
 		}
-		nladdr.nl_pid = 0;
+		nladdr.nl_pid = 0; /* XXX clang static analyzer bug? */
 		nladdr_len = sizeof(nladdr);
 		bytes = recvfrom(fd, buf, buflen, flags,
 		    (struct sockaddr *)&nladdr, &nladdr_len);
@@ -348,16 +385,15 @@ get_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp, int fd, int flags,
 			continue;
 		}
 
-		for (nlm = (struct nlmsghdr *)(void *)buf;
+		for (nlm = (void *)buf;
 		     nlm && NLMSG_OK(nlm, (size_t)bytes);
 		     nlm = NLMSG_NEXT(nlm, bytes))
 		{
 			r = err_netlink(nlm);
 			if (r == -1)
 				goto eexit;
-			if (r)
-				continue;
-			if (callback) {
+			
+			if (r == 0 && callback) {
 				r = callback(ctx, ifp, nlm);
 				if (r != 0)
 					goto eexit;
@@ -377,6 +413,7 @@ if_copyrt(struct dhcpcd_ctx *ctx, struct rt *rt, struct nlmsghdr *nlm)
 	size_t len;
 	struct rtmsg *rtm;
 	struct rtattr *rta;
+	unsigned int ifindex;
 
 	len = nlm->nlmsg_len - sizeof(*nlm);
 	if (len < sizeof(*rtm)) {
@@ -410,8 +447,8 @@ if_copyrt(struct dhcpcd_ctx *ctx, struct rt *rt, struct nlmsghdr *nlm)
 			    sizeof(rt->src.s_addr));
 			break;
 		case RTA_OIF:
-			rt->iface = if_findindex(ctx->ifaces,
-			    *(unsigned int *)RTA_DATA(rta));
+			ifindex = *(unsigned int *)RTA_DATA(rta);
+			rt->iface = if_findindex(ctx->ifaces, ifindex);
 			break;
 		case RTA_PRIORITY:
 			rt->metric = *(unsigned int *)RTA_DATA(rta);
@@ -448,6 +485,11 @@ if_copyrt(struct dhcpcd_ctx *ctx, struct rt *rt, struct nlmsghdr *nlm)
 		if ((ap = ipv4_findaddr(ctx, &rt->src)))
 			rt->iface = ap->iface;
 	}
+
+	if (rt->iface == NULL) {
+		errno = ESRCH;
+		return -1;
+	}
 	return 0;
 }
 #endif
@@ -459,6 +501,7 @@ if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct nlmsghdr *nlm)
 	size_t len;
 	struct rtmsg *rtm;
 	struct rtattr *rta;
+	unsigned int ifindex;
 
 	len = nlm->nlmsg_len - sizeof(*nlm);
 	if (len < sizeof(*rtm)) {
@@ -489,8 +532,8 @@ if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct nlmsghdr *nlm)
 			    sizeof(rt->gate.s6_addr));
 			break;
 		case RTA_OIF:
-			rt->iface = if_findindex(ctx->ifaces,
-			    *(unsigned int *)RTA_DATA(rta));
+			ifindex = *(unsigned int *)RTA_DATA(rta);
+			rt->iface = if_findindex(ctx->ifaces, ifindex);
 			break;
 		case RTA_PRIORITY:
 			rt->metric = *(unsigned int *)RTA_DATA(rta);
@@ -516,22 +559,14 @@ if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct nlmsghdr *nlm)
 		rta = RTA_NEXT(rta, len);
 	}
 
+	if (rt->iface == NULL) {
+		errno = ESRCH;
+		return -1;
+	}
 	return 0;
 }
 #endif
 
-/* Work out the maximum pid size */
-static inline long long
-get_max_pid_t()
-{
-
-	if (sizeof(pid_t) == sizeof(short))		return SHRT_MAX;
-	if (sizeof(pid_t) == sizeof(int))		return INT_MAX;
-	if (sizeof(pid_t) == sizeof(long))		return LONG_MAX;
-	if (sizeof(pid_t) == sizeof(long long))		return LLONG_MAX;
-	abort();
-}
-
 static int
 link_route(struct dhcpcd_ctx *ctx, __unused struct interface *ifp,
     struct nlmsghdr *nlm)
@@ -539,6 +574,7 @@ link_route(struct dhcpcd_ctx *ctx, __unused struct interface *ifp,
 	size_t len;
 	struct rtmsg *rtm;
 	int cmd;
+	struct priv *priv;
 #ifdef INET
 	struct rt rt;
 #endif
@@ -562,19 +598,17 @@ link_route(struct dhcpcd_ctx *ctx, __unused struct interface *ifp,
 		return -1;
 	}
 
-	/* Ignore messages generated by us.
-	 * For some reason we get messages generated by us
-	 * with a very large value in nlmsg_pid that seems to be
-	 * sequentially changing. Is there a better test for this? */
-	if (nlm->nlmsg_pid > get_max_pid_t())
-		return 1;
+	/* Ignore messages we sent. */
+	priv = (struct priv *)ctx->priv;
+	if (nlm->nlmsg_pid == priv->route_pid)
+		return 0;
 
 	rtm = NLMSG_DATA(nlm);
 	switch (rtm->rtm_family) {
 #ifdef INET
 	case AF_INET:
 		if (if_copyrt(ctx, &rt, nlm) == 0)
-			ipv4_handlert(ctx, cmd, &rt);
+			ipv4_handlert(ctx, cmd, &rt, 0);
 		break;
 #endif
 #ifdef INET6
@@ -594,6 +628,7 @@ link_addr(struct dhcpcd_ctx *ctx, struct interface *ifp, struct nlmsghdr *nlm)
 	size_t len;
 	struct rtattr *rta;
 	struct ifaddrmsg *ifa;
+	struct priv *priv;
 #ifdef INET
 	struct in_addr addr, net, dest;
 #endif
@@ -610,12 +645,10 @@ link_addr(struct dhcpcd_ctx *ctx, struct interface *ifp, struct nlmsghdr *nlm)
 		return -1;
 	}
 
-	/* Ignore messages generated by us.
-	 * For some reason we get messages generated by us
-	 * with a very large value in nlmsg_pid that seems to be
-	 * sequentially changing. Is there a better test for this? */
-	if (nlm->nlmsg_pid > get_max_pid_t())
-		return 1;
+	/* Ignore messages we sent. */
+	priv = (struct priv*)ctx->priv;
+	if (nlm->nlmsg_pid == priv->route_pid)
+		return 0;
 
 	ifa = NLMSG_DATA(nlm);
 	if ((ifp = if_findindex(ctx->ifaces, ifa->ifa_index)) == NULL) {
@@ -783,7 +816,7 @@ link_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp,
 	ifi = NLMSG_DATA(nlm);
 	if (ifi->ifi_flags & IFF_LOOPBACK)
 		return 0;
-	rta = (struct rtattr *)(void *)((char *)ifi +NLMSG_ALIGN(sizeof(*ifi)));
+	rta = (void *)((char *)ifi + NLMSG_ALIGN(sizeof(*ifi)));
 	len = NLMSG_PAYLOAD(nlm, sizeof(*ifi));
 	*ifn = '\0';
 	hwaddr = NULL;
@@ -865,28 +898,40 @@ send_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp,
 	struct sockaddr_nl snl;
 	struct iovec iov;
 	struct msghdr msg;
-	static unsigned int seq;
 
 	memset(&snl, 0, sizeof(snl));
-	if ((s = _open_link_socket(&snl, protocol)) == -1)
-		return -1;
-	memset(&iov, 0, sizeof(iov));
-	iov.iov_base = hdr;
-	iov.iov_len = hdr->nlmsg_len;
+	snl.nl_family = AF_NETLINK;
+
+	if (protocol == NETLINK_ROUTE) {
+		struct priv *priv;
+
+		priv = (struct priv *)ctx->priv;
+		s = priv->route_fd;
+	} else {
+		if ((s = _open_link_socket(&snl, protocol)) == -1)
+			return -1;
+	}
+
 	memset(&msg, 0, sizeof(msg));
 	msg.msg_name = &snl;
 	msg.msg_namelen = sizeof(snl);
+	memset(&iov, 0, sizeof(iov));
+	iov.iov_base = hdr;
+	iov.iov_len = hdr->nlmsg_len;
 	msg.msg_iov = &iov;
 	msg.msg_iovlen = 1;
 	/* Request a reply */
 	hdr->nlmsg_flags |= NLM_F_ACK;
-	hdr->nlmsg_seq = ++seq;
-
-	if (sendmsg(s, &msg, 0) != -1)
+	hdr->nlmsg_seq = (uint32_t)++ctx->seq;
+	if ((unsigned int)ctx->seq > UINT32_MAX)
+		ctx->seq = 0;
+	if (sendmsg(s, &msg, 0) != -1) {
+		ctx->sseq = ctx->seq;
 		r = get_netlink(ctx, ifp, s, 0, callback);
-	else
+	} else
 		r = -1;
-	close(s);
+	if (protocol != NETLINK_ROUTE)
+		close(s);
 	return r;
 }
 
@@ -948,8 +993,7 @@ rta_add_attr_32(struct rtattr *rta, unsigned short maxlen,
 		return -1;
 	}
 
-	subrta = (struct rtattr*)(void *)
-	    (((char*)rta) + RTA_ALIGN(rta->rta_len));
+	subrta = (void *)((char*)rta + RTA_ALIGN(rta->rta_len));
 	subrta->rta_type = type;
 	subrta->rta_len = len;
 	memcpy(RTA_DATA(subrta), &data, sizeof(data));
@@ -962,8 +1006,8 @@ static struct nlattr *
 nla_next(struct nlattr *nla, size_t *rem)
 {
 
-	*rem -= NLA_ALIGN(nla->nla_len);
-	return (struct nlattr *)(void *)((char *)nla + NLA_ALIGN(nla->nla_len));
+	*rem -= (size_t)NLA_ALIGN(nla->nla_len);
+	return (void *)((char *)nla + NLA_ALIGN(nla->nla_len));
 }
 
 #define NLA_TYPE(nla) ((nla)->nla_type & NLA_TYPE_MASK)
@@ -1037,7 +1081,7 @@ gnl_parse(struct nlmsghdr *nlm, struct nlattr *tb[], int maxtype)
 
 	memset(tb, 0, sizeof(*tb) * ((unsigned int)maxtype + 1));
 	ghdr = NLMSG_DATA(nlm);
-	head = (struct nlattr *)(void *)((char *) ghdr + GENL_HDRLEN);
+	head = (void *)((char *)ghdr + GENL_HDRLEN);
 	len = nlm->nlmsg_len - GENL_HDRLEN - NLMSG_HDRLEN;
 	NLA_FOR_EACH_ATTR(nla, head, len, rem) {
 		type = NLA_TYPE(nla);
@@ -1100,7 +1144,7 @@ _if_getssid(__unused struct dhcpcd_ctx *ctx, struct interface *ifp,
 	}
 
 	ifp->ssid_len = NLA_LEN(tb[NL80211_ATTR_SSID]);
-	if (ifp->ssid_len > sizeof(ifp->ssid)) {
+	if (ifp->ssid_len > IF_SSIDLEN) {
 		errno = ENOBUFS;
 		ifp->ssid_len = 0;
 		return -1;
@@ -1108,7 +1152,6 @@ _if_getssid(__unused struct dhcpcd_ctx *ctx, struct interface *ifp,
 	memcpy(ifp->ssid, NLA_DATA(tb[NL80211_ATTR_SSID]), ifp->ssid_len);
 
 out:
-	ifp->ssid[ifp->ssid_len] = '\0';
 	return (int)ifp->ssid_len;
 }
 
@@ -1142,10 +1185,16 @@ if_getssid(struct interface *ifp)
 	r = if_getssid_wext(ifp->name, ifp->ssid);
 	if (r != -1)
 		ifp->ssid_len = (unsigned int)r;
+	else {
 #ifdef HAVE_NL80211_H
-	else if (r == -1)
 		r = if_getssid_nl80211(ifp);
+		if (r == -1)
+			ifp->ssid_len = 0;
+#else
+		ifp->ssid_len = 0;
 #endif
+	}
+	ifp->ssid[ifp->ssid_len] = '\0';
 	return r;
 }
 
@@ -1180,9 +1229,10 @@ if_openrawsocket(struct interface *ifp, uint16_t protocol)
 	int n;
 #endif
 
-	if ((s = xsocket(PF_PACKET, SOCK_DGRAM, htons(protocol),
-	    O_CLOEXEC | O_NONBLOCK)) == -1)
+#define SF	SOCK_CLOEXEC | SOCK_NONBLOCK
+	if ((s = xsocket(PF_PACKET, SOCK_DGRAM, htons(protocol), SF)) == -1)
 		return -1;
+#undef SF
 
 	/* Install the DHCP filter */
 	memset(&pf, 0, sizeof(pf));
@@ -1378,9 +1428,6 @@ if_route(unsigned char cmd, const struct rt *rt)
 		if (rt->gate.s_addr != htonl(INADDR_ANY))
 			add_attr_l(&nlm.hdr, sizeof(nlm), RTA_GATEWAY,
 			    &rt->gate.s_addr, sizeof(rt->gate.s_addr));
-		if (rt->gate.s_addr != htonl(INADDR_LOOPBACK))
-			add_attr_32(&nlm.hdr, sizeof(nlm), RTA_OIF,
-			    rt->iface->index);
 		if (subnet != -1) {
 			add_attr_l(&nlm.hdr, sizeof(nlm), RTA_PREFSRC,
 			    &src_addr.s_addr, sizeof(src_addr.s_addr));
@@ -1399,6 +1446,9 @@ if_route(unsigned char cmd, const struct rt *rt)
 		}
 	}
 
+	if (rt->gate.s_addr != htonl(INADDR_LOOPBACK))
+		add_attr_32(&nlm.hdr, sizeof(nlm), RTA_OIF, rt->iface->index);
+
 	if (rt->metric)
 		add_attr_32(&nlm.hdr, sizeof(nlm), RTA_PRIORITY, rt->metric);
 
@@ -1413,7 +1463,7 @@ _if_initrt(struct dhcpcd_ctx *ctx, __unused struct interface *ifp,
 	struct rt rt;
 
 	if (if_copyrt(ctx, &rt, nlm) == 0)
-		ipv4_handlert(ctx, RTM_ADD, &rt);
+		ipv4_handlert(ctx, RTM_ADD, &rt, 1);
 	return 0;
 }
 
