@@ -1,6 +1,6 @@
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2015 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2016 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -30,12 +30,11 @@
 #include <sys/ioctl.h>
 #include <sys/socket.h>
 
+#include "config.h"
+
 #include <net/if.h>
 #include <net/if_arp.h>
 #include <netinet/in.h>
-#ifdef __FreeBSD__ /* Needed so that including netinet6/in6_var.h works */
-#  include <net/if_var.h>
-#endif
 #ifdef AF_LINK
 #  include <net/if_dl.h>
 #  include <net/if_types.h>
@@ -60,7 +59,6 @@
 #include <unistd.h>
 #include <fcntl.h>
 
-#include "config.h"
 #include "common.h"
 #include "dev.h"
 #include "dhcp.h"
@@ -91,21 +89,16 @@ int
 if_opensockets(struct dhcpcd_ctx *ctx)
 {
 
-	if ((ctx->link_fd = if_openlinksocket()) == -1)
+	if (if_opensockets_os(ctx) == -1)
 		return -1;
 
-	ctx->pf_inet_fd = xsocket(PF_INET, SOCK_DGRAM, 0, O_CLOEXEC);
+	/* We use this socket for some operations without INET. */
+	ctx->pf_inet_fd = xsocket(PF_INET, SOCK_DGRAM, 0, SOCK_CLOEXEC);
 	if (ctx->pf_inet_fd == -1)
 		return -1;
 
-#if defined(INET6) && defined(BSD)
-	ctx->pf_inet6_fd = xsocket(PF_INET6, SOCK_DGRAM, 0, O_CLOEXEC);
-	if (ctx->pf_inet6_fd == -1)
-		return -1;
-#endif
-
 #ifdef IFLR_ACTIVE
-	ctx->pf_link_fd = xsocket(PF_LINK, SOCK_DGRAM, 0, O_CLOEXEC);
+	ctx->pf_link_fd = xsocket(PF_LINK, SOCK_DGRAM, 0, SOCK_CLOEXEC);
 	if (ctx->pf_link_fd == -1)
 		return -1;
 #endif
@@ -113,6 +106,23 @@ if_opensockets(struct dhcpcd_ctx *ctx)
 	return 0;
 }
 
+void
+if_closesockets(struct dhcpcd_ctx *ctx)
+{
+
+	if (ctx->pf_inet_fd != -1)
+		close(ctx->pf_inet_fd);
+#ifdef IFLR_ACTIVE
+	if (ctx->pf_link_fd != -1)
+		close(ctx->pf_link_fd);
+#endif
+
+	if (ctx->priv) {
+		if_closesockets_os(ctx);
+		free(ctx->priv);
+	}
+}
+
 int
 if_carrier(struct interface *ifp)
 {
@@ -176,7 +186,7 @@ if_hasconf(struct dhcpcd_ctx *ctx, const char *ifname)
 	return 0;
 }
 
-static void if_learnaddrs1(struct dhcpcd_ctx *ctx, struct if_head *ifs,
+static void if_learnaddrs(struct dhcpcd_ctx *ctx, struct if_head *ifs,
     struct ifaddrs *ifaddrs)
 {
 	struct ifaddrs *ifa;
@@ -198,10 +208,8 @@ static void if_learnaddrs1(struct dhcpcd_ctx *ctx, struct if_head *ifs,
 		switch(ifa->ifa_addr->sa_family) {
 #ifdef INET
 		case AF_INET:
-			addr = (const struct sockaddr_in *)
-			    (void *)ifa->ifa_addr;
-			net = (const struct sockaddr_in *)
-			    (void *)ifa->ifa_netmask;
+			addr = (void *)ifa->ifa_addr;
+			net = (void *)ifa->ifa_netmask;
 			if (ifa->ifa_flags & IFF_POINTOPOINT)
 				dst = (const struct sockaddr_in *)
 				    (void *)ifa->ifa_dstaddr;
@@ -216,8 +224,8 @@ static void if_learnaddrs1(struct dhcpcd_ctx *ctx, struct if_head *ifs,
 #endif
 #ifdef INET6
 		case AF_INET6:
-			sin6 = (struct sockaddr_in6 *)(void *)ifa->ifa_addr;
-			net6 = (struct sockaddr_in6 *)(void *)ifa->ifa_netmask;
+			sin6 = (void *)ifa->ifa_addr;
+			net6 = (void *)ifa->ifa_netmask;
 #ifdef __KAME__
 			if (IN6_IS_ADDR_LINKLOCAL(&sin6->sin6_addr))
 				/* Remove the scope from the address */
@@ -242,7 +250,7 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 {
 	struct ifaddrs *ifaddrs, *ifa;
 	char *p;
-	int i;
+	int i, active;
 	struct if_head *ifs;
 	struct interface *ifp;
 #ifdef __linux__
@@ -291,6 +299,7 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 		if (ifp)
 			continue;
 
+		active = IF_ACTIVE_USER;
 		if (argc > 0) {
 			for (i = 0; i < argc; i++) {
 #ifdef __linux__
@@ -306,9 +315,14 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 					break;
 #endif
 			}
-			if (i == argc)
-				continue;
-			p = argv[i];
+			if (i == argc) {
+				active = IF_INACTIVE;
+				p =  ifa->ifa_name;
+#ifdef __linux__
+				strlcpy(ifn, ifa->ifa_name, sizeof(ifn));
+#endif
+			} else
+				p = argv[i];
 		} else {
 			p = ifa->ifa_name;
 #ifdef __linux__
@@ -320,16 +334,17 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 			if (argc == -1 && strcmp(argv[0], ifa->ifa_name) != 0)
 				continue;
 		}
+
 		for (i = 0; i < ctx->ifdc; i++)
 			if (!fnmatch(ctx->ifdv[i], p, 0))
 				break;
 		if (i < ctx->ifdc)
-			continue;
+			active = IF_INACTIVE;
 		for (i = 0; i < ctx->ifac; i++)
 			if (!fnmatch(ctx->ifav[i], p, 0))
 				break;
 		if (ctx->ifac && i == ctx->ifac)
-			continue;
+			active = IF_INACTIVE;
 
 #ifdef PLUGIN_DEV
 		/* Ensure that the interface name has settled */
@@ -341,7 +356,7 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 		if (ifa->ifa_flags & (IFF_LOOPBACK | IFF_POINTOPOINT)) {
 			if ((argc == 0 || argc == -1) &&
 			    ctx->ifac == 0 && !if_hasconf(ctx, p))
-				continue;
+				active = IF_INACTIVE;
 		}
 
 		if (if_vimaster(ctx, p) == 1) {
@@ -367,7 +382,7 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 
 		if (ifa->ifa_addr != NULL) {
 #ifdef AF_LINK
-			sdl = (const struct sockaddr_dl *)(void *)ifa->ifa_addr;
+			sdl = (const void *)ifa->ifa_addr;
 
 #ifdef IFLR_ACTIVE
 			/* We need to check for active address */
@@ -399,7 +414,7 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 #if defined(IFT_BRIDGE) || defined(IFT_PPP) || defined(IFT_PROPVIRTUAL)
 				/* Don't allow unless explicit */
 				if ((argc == 0 || argc == -1) &&
-				    ctx->ifac == 0 &&
+				    ctx->ifac == 0 && active &&
 				    !if_hasconf(ctx, ifp->name))
 				{
 					logger(ifp->ctx, LOG_DEBUG,
@@ -407,8 +422,7 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 					    " interface type and"
 					    " no config",
 					    ifp->name);
-					if_free(ifp);
-					continue;
+					active = IF_INACTIVE;
 				}
 				/* FALLTHROUGH */
 #endif
@@ -436,13 +450,12 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 				if ((argc == 0 || argc == -1) &&
 				    ctx->ifac == 0 &&
 				    !if_hasconf(ctx, ifp->name))
-				{
-					if_free(ifp);
-					continue;
-				}
-				logger(ifp->ctx, LOG_WARNING,
-				    "%s: unsupported interface type %.2x",
-				    ifp->name, sdl->sdl_type);
+					active = IF_INACTIVE;
+				if (active)
+					logger(ifp->ctx, LOG_WARNING,
+					    "%s: unsupported"
+					    " interface type %.2x",
+					    ifp->name, sdl->sdl_type);
 				/* Pretend it's ethernet */
 				ifp->family = ARPHRD_ETHER;
 				break;
@@ -453,7 +466,7 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 #endif
 			memcpy(ifp->hwaddr, CLLADDR(sdl), ifp->hwlen);
 #elif AF_PACKET
-			sll = (const struct sockaddr_ll *)(void *)ifa->ifa_addr;
+			sll = (const void *)ifa->ifa_addr;
 			ifp->index = (unsigned int)sll->sll_ifindex;
 			ifp->family = sll->sll_hatype;
 			ifp->hwlen = sll->sll_halen;
@@ -471,10 +484,7 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 		if (ifp->family != ARPHRD_ETHER) {
 			if ((argc == 0 || argc == -1) &&
 			    ctx->ifac == 0 && !if_hasconf(ctx, ifp->name))
-			{
-				if_free(ifp);
-				continue;
-			}
+				active = IF_INACTIVE;
 			switch (ifp->family) {
 			case ARPHRD_IEEE1394:
 			case ARPHRD_INFINIBAND:
@@ -490,9 +500,11 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 /* IFT already checked */
 #ifndef AF_LINK
 			default:
-				logger(ifp->ctx, LOG_WARNING,
-				    "%s: unsupported interface family %.2x",
-				    ifp->name, ifp->family);
+				if (active)
+					logger(ifp->ctx, LOG_WARNING,
+					    "%s: unsupported"
+					    " interface family %.2x",
+					    ifp->name, ifp->family);
 				break;
 #endif
 			}
@@ -500,14 +512,14 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 
 		if (!(ctx->options & (DHCPCD_DUMPLEASE | DHCPCD_TEST))) {
 			/* Handle any platform init for the interface */
-			if (if_init(ifp) == -1) {
+			if (active && if_init(ifp) == -1) {
 				logger(ifp->ctx, LOG_ERR, "%s: if_init: %m", p);
 				if_free(ifp);
 				continue;
 			}
 
 			/* Ensure that the MTU is big enough for DHCP */
-			if (if_getmtu(ifp) < MTU_MIN &&
+			if (if_getmtu(ifp) < MTU_MIN && active &&
 			    if_setmtu(ifp, MTU_MIN) == -1)
 			{
 				logger(ifp->ctx, LOG_ERR,
@@ -523,6 +535,7 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 		strlcpy(ifr.ifr_name, ifp->name, sizeof(ifr.ifr_name));
 		if (ioctl(ctx->pf_inet_fd, SIOCGIFPRIORITY, &ifr) == 0)
 			ifp->metric = (unsigned int)ifr.ifr_metric;
+		if_getssid(ifp);
 #else
 		/* We reserve the 100 range for virtual interfaces, if and when
 		 * we can work them out. */
@@ -533,10 +546,11 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 		}
 #endif
 
+		ifp->active = active;
 		TAILQ_INSERT_TAIL(ifs, ifp, next);
 	}
 
-	if_learnaddrs1(ctx, ifs, ifaddrs);
+	if_learnaddrs(ctx, ifs, ifaddrs);
 	freeifaddrs(ifaddrs);
 
 	return ifs;
@@ -559,7 +573,7 @@ if_findindexname(struct if_head *ifaces, unsigned int idx, const char *name)
 		}
 	}
 
-	errno = ESRCH;
+	errno = ENXIO;
 	return NULL;
 }
 
@@ -600,7 +614,13 @@ if_cmp(const struct interface *si, const struct interface *ti)
 	int r;
 #endif
 
-	/* Check carrier status first */
+	/* Check active first */
+	if (si->active > ti->active)
+		return -1;
+	if (si->active < ti->active)
+		return 1;
+
+	/* Check carrier status next */
 	if (si->carrier > ti->carrier)
 		return -1;
 	if (si->carrier < ti->carrier)
@@ -665,25 +685,37 @@ if_sortinterfaces(struct dhcpcd_ctx *ctx)
 int
 xsocket(int domain, int type, int protocol, int flags)
 {
-#ifdef SOCK_CLOEXEC
-	if (flags & O_CLOEXEC)
-		type |= SOCK_CLOEXEC;
-	if (flags & O_NONBLOCK)
-		type |= SOCK_NONBLOCK;
+	int s;
+#if !defined(HAVE_SOCK_CLOEXEC) || !defined(HAVE_SOCK_NONBLOCK)
+	int xflags;
+#endif
 
-	return socket(domain, type, protocol);
-#else
-	int s, xflags;
+#ifdef HAVE_SOCK_CLOEXEC
+	if (flags & SOCK_CLOEXEC)
+		type |= SOCK_CLOEXEC;
+#endif
+#ifdef HAVE_SOCK_NONBLOCK
+	if (flags & SOCK_NONBLOCK)
+		type |= SOCK_NONBLOCK;
+#endif
 
 	if ((s = socket(domain, type, protocol)) == -1)
 		return -1;
-	if ((flags & O_CLOEXEC) && ((xflags = fcntl(s, F_GETFD, 0)) == -1 ||
+
+#ifndef HAVE_SOCK_CLOEXEC
+	if ((flags & SOCK_CLOEXEC) && ((xflags = fcntl(s, F_GETFD)) == -1 ||
 	    fcntl(s, F_SETFD, xflags | FD_CLOEXEC) == -1))
 		goto out;
-	if ((flags & O_NONBLOCK) && ((xflags = fcntl(s, F_GETFL, 0)) == -1 ||
+#endif
+#ifndef HAVE_SOCK_NONBLOCK
+	if ((flags & SOCK_NONBLOCK) && ((xflags = fcntl(s, F_GETFL)) == -1 ||
 	    fcntl(s, F_SETFL, xflags | O_NONBLOCK) == -1))
 		goto out;
+#endif
+
 	return s;
+
+#if !defined(HAVE_SOCK_CLOEXEC) || !defined(HAVE_SOCK_NONBLOCK)
 out:
 	close(s);
 	return -1;
