@@ -176,9 +176,12 @@ ipv4ll_probed(struct arp_state *astate)
 #endif
 		logger(ifp->ctx, LOG_INFO, "%s: using IPv4LL address %s",
 		  ifp->name, inet_ntoa(astate->addr));
-	if (ia == NULL)
+	if (ia == NULL) {
+		if (ifp->ctx->options & DHCPCD_TEST)
+			goto test;
 		ia = ipv4_addaddr(ifp, &astate->addr,
 		    &inaddr_llmask, &inaddr_llbcast);
+	}
 	if (ia == NULL)
 		return;
 #ifdef IN_IFF_NOTREADY
@@ -187,7 +190,13 @@ ipv4ll_probed(struct arp_state *astate)
 	logger(ifp->ctx, LOG_DEBUG, "%s: DAD completed for %s",
 	    ifp->name, inet_ntoa(astate->addr));
 #endif
+test:
 	state->addr = astate->addr;
+	if (ifp->ctx->options & DHCPCD_TEST) {
+		script_runreason(ifp, "TEST");
+		eloop_exit(ifp->ctx->eloop, EXIT_SUCCESS);
+		return;
+	}
 	timespecclear(&state->defend);
 	if_initrt(ifp);
 	ipv4_buildroutes(ifp->ctx);
@@ -435,3 +444,29 @@ ipv4ll_freedrop(struct interface *ifp, int drop)
 		}
 	}
 }
+
+/* This may cause issues in BSD systems, where running as a single dhcpcd
+ * daemon would solve this issue easily. */
+#ifdef HAVE_ROUTE_METRIC
+int
+ipv4ll_handlert(struct dhcpcd_ctx *ctx, __unused int cmd, const struct rt *rt)
+{
+	struct interface *ifp;
+
+	/* Only interested in default route changes. */
+	if (rt->dest.s_addr != INADDR_ANY)
+		return 0;
+
+	/* If any interface is running IPv4LL, rebuild our routing table. */
+	TAILQ_FOREACH(ifp, ctx->ifaces, next) {
+		if (IPV4LL_STATE_RUNNING(ifp))
+			break;
+	}
+	if (ifp != NULL) {
+		if_initrt(ifp);
+		ipv4_buildroutes(ctx);
+	}
+
+	return 0;
+}
+#endif
